"""
This module contains the classes to create a plot(s) window.

@author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""

import queue
import sys
from PyQt5 import QtCore, QtWidgets
from data_dialogs import AxisDialog
import pyqtgraph as pg

class MainWindow(QtWidgets.QMainWindow):
    """This class wraps the PlotWindow class to add extra functionality."""

    def __init__(self, data_queue, plots, plot_names, beh_intervals=None, parent=None):
        super().__init__(parent)

        # Create plot window widget & set it as the central widget
        self.plot_window = PlotWindow(data_queue, plots, plot_names, beh_intervals=beh_intervals, parent=self)
        self.setCentralWidget(self.plot_window)

        self.statusbar = self.statusBar()
        self.statusbar.showMessage("Ready")

        menu_bar = self.menuBar()

        # File menu
        file_menu = menu_bar.addMenu("File")

        # Create submenu to start new processes from file menu
        adjust_plots = QtWidgets.QMenu("Adjust plots", self)
        file_menu.addMenu(adjust_plots)
        set_x_axis = QtWidgets.QAction("set x axis", self)
        set_y_axis = QtWidgets.QAction("set y axis", self)
        adjust_plots.addAction(set_x_axis)
        adjust_plots.addAction(set_y_axis)
        set_x_axis.triggered.connect(self.on_set_axis)
        set_y_axis.triggered.connect(self.on_set_y_axis)

        # Create and connect action to close the plot window
        close_action = QtWidgets.QAction("Close", self)
        file_menu.addAction(close_action)
        close_action.triggered.connect(sys.exit)

    def on_set_axis(self, axis=0):
        """Handles both actions in the 'Adjust plots' submenu"""

        lower_bound, upper_bound = show_axis_dialog()

        if lower_bound and upper_bound:

            # lineedit returns str, so first convert both to int
            lower_bound = int(lower_bound)
            upper_bound = int(upper_bound)

            for plot_item in self.plot_window.plot_items:
                if axis:
                    plot_item.setRange(yRange=[lower_bound, upper_bound], padding=0)
                else:
                    plot_item.setRange(xRange=[lower_bound, upper_bound], padding=0)

    def on_set_y_axis(self):
        self.on_set_axis(axis=1)


def show_axis_dialog():
    """Display the behavior dialog to the user & let user choose colors.
    """
    # app = QtWidgets.QApplication(sys.argv)
    axis_dialog = AxisDialog()
    axis_dialog.exec_()
    axis_dialog.show()
    return axis_dialog.lower_bound, axis_dialog.upper_bound


class PlotWindow(pg.GraphicsWindow):
    """This class holds plots and should be nested in a MainWindow."""

    changed_behavior = QtCore.pyqtSignal('QString')

    def __init__(self, data_queue, plots, plot_names, beh_intervals=None, parent=None):
        super().__init__(parent=parent)
        self.parent = parent

        self.graphics_layout = pg.GraphicsLayout(border=(0, 0, 0))
        self.graphics_layout.layout.setSpacing(0)
        self.graphics_layout.layout.setContentsMargins(0, 0, 0, 0)
        self.setCentralItem(self.graphics_layout)

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(10)
        self.timer.timeout.connect(self.on_new_data)

        self.behavior_time = {}
        self.plot_items = []
        self.vertical_lines = []
        self.plots = plots
        self.add_plots(plot_names, beh_intervals)

        # Convert all sets to strings
        if beh_intervals:
            self.behavior_time = {time: ', '.join(behavior) for time, behavior in self.behavior_time.items()}
        else:
            self.behavior_time = None

        self.data_queue = data_queue
        self.timer.start()

    def on_new_data(self):
        """Called on timer interval to get data from queue & update all plots.
        """
        try:
            val = self.data_queue.get(block=False)
        except queue.Empty:
            QtCore.QCoreApplication.processEvents()
            return

        if val in ('<', '>', 'P', 'p'):
            return
        if val == 'S':
            for v_line in self.vertical_lines:
                v_line.setValue(0)
            if self.behavior_time:
                self.parent.statusbar.showMessage(self.behavior_time[0])
            return

        val = int(val)
        val //= 100
        for v_line in self.vertical_lines:
            v_line.setValue(val)
        if self.behavior_time:
            self.parent.statusbar.showMessage(self.behavior_time[val])

        QtCore.QCoreApplication.processEvents()

    def add_plots(self, plot_names, all_beh_intervals):
        """Creates and set all the plots"""

        # Find the maximum value of the x-axis for all plots
        x_max = len(self.plots[0])

        # Pen to be used for vertical lines
        pen = pg.mkPen('r', width=2)

        # Get background color(s) (color coded by behavior) to add to each plot
        list_of_lists_of_rgns = [[] for _ in self.plots]
        if all_beh_intervals:
            for behavior_intervals, behavior, color in all_beh_intervals:
                for interval in behavior_intervals:

                    # Add the interval times and the corresponding behavior
                    # to track it for the status bar
                    for time in interval:
                        if time not in self.behavior_time:
                            self.behavior_time[time] = {behavior}
                        else:
                            self.behavior_time[time].add(behavior)

                    # Add rgn object for each plot
                    # (Since copies are not possible)
                    for inner_lst in list_of_lists_of_rgns:
                        rgn = pg.LinearRegionItem(values=[interval[0], interval[-1]], movable=False)
                        rgn.lines[0].setPen((255, 255, 255, 5))
                        rgn.lines[1].setPen((255, 255, 255, 5))
                        rgn.setBrush(pg.mkBrush(color))
                        inner_lst.append(rgn)

        for i, plot in enumerate(self.plots):
            plot_item = self.graphics_layout.addPlot(title="{}".format(plot_names[i]), row=i, col=0)

            # Causes auto-scale button (‘A’ in lower-left corner)
            # to be hidden for this PlotItem
            plot_item.hideButtons()

            plot_item.setMouseEnabled(x=True, y=False)

            for rgn in list_of_lists_of_rgns[i]:
                plot_item.addItem(rgn)

            plot_item.plot(plot, pen=pg.mkPen((0, 0, 0), width=2))

            # Get the max value in the time series to set the yRange, below
            y_max = plot.max()

            # Set the domain and range for each plot
            plot_item.setRange(xRange=[0, x_max], yRange=[-1, y_max], padding=0)

            # Create and add vertical line that scrolls
            self.vertical_line = plot_item.addLine(x=0, pen=pen)
            self.vertical_lines.append(self.vertical_line)

            # Store references to each PlotItem
            self.plot_items.append(plot_item)


if __name__ == "__main__":
    pg.setConfigOption("background", 'w')
    pg.setConfigOption("foreground", 'k')



"""
This module contains useful functions for clustering.

@authors: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
          Jack Berry, Columbia University, jeb2242@columbia.edu
"""


from sklearn import cluster
from sklearn.metrics import normalized_mutual_info_score
from sklearn.metrics import mutual_info_score
import itertools
import pandas as pd
import numpy as np

def similarity_matrix(dataframe, similarity_measure=normalized_mutual_info_score):
    """
    Create a DataFrame similarity matrix

    Inputs:
        dataframe: pandas DataFrame
            A TxN DataFrame where T is the number of observations (time bins)
            and N is the number of variables (neurons)

        similarity_measure: function, optional
            The similarity metric to be used. Default is
            sklearn.metrics.normalized_mutual_info_score. any function can be used,
            including the pearsonr wrapper in this module.

    Returns:
        matrix: pandas DataFrame
            A DataFrame storing the similarity matrix for all neurons
    """
    rows = []

    for col in dataframe:
        row = dataframe.apply(similarity_measure, args=(dataframe[col],))
        rows.append(row)

    matrix = pd.DataFrame(rows, index=dataframe.columns)
    return matrix

def compute_connections(dataframe, similarity_measure=normalized_mutual_info_score):
    """
    Create a dict of neuron pairs and their corresponding correlation/similarity

    Inputs:
        dataframe: pandas DataFrame
            A TxN DataFrame where T is the number of observations (time bins)
            and N is the number of variables (neurons)

        similarity_measure: function, optional
            The similarity metric to be used. Default is
            sklearn.metrics.normalized_mutual_info_score. any function can be
            used, including the pearsonr wrapper in this module.

    Returns:
        connections: dict
            A dictionary in the form of {(i,j):similarity}
    """


    connections = {}
    matrix = similarity_matrix(dataframe, similarity_measure)

    for combination in itertools.combinations(matrix.columns, 2):
        similarity = matrix[combination[0]][combination[1]]
        connections[combination] = similarity

    return connections

def pearsonr(a, b):
    """
    Wrapper for the numpy corrcoef function to compute pearson correlation coefficient.

    Inputs:
        a,b: array-like
            vectors to compute pearson correlation coefficient

    Returns:
        corr: float
            The pearson correlation coefficient

    """
    corr = np.corrcoef(a, b)[0][1]
    return corr

def compute_nmi(dataframe):
    connections = {}
    matrix = nmi_matrix(dataframe)

    for combination in itertools.combinations(matrix.columns, 2):
        nmi = matrix[combination[0]][combination[1]]
        connections[combination] = nmi

    return connections

def nmi_matrix(dataframe):
    rows = []

    for col in dataframe:
        row = dataframe.apply(normalized_mutual_info_score, args=(dataframe[col],))
        rows.append(row)

    return pd.DataFrame(rows, index=dataframe.columns)

def compute_mi(dataframe):
    connections = {}
    matrix = mi_matrix(dataframe)

    for combination in itertools.combinations(matrix.columns, 2):
        mi = matrix[combination[0]][combination[1]]
        connections[combination] = mi

    return connections

def mi_matrix(dataframe):
    rows = []

    for col in dataframe:
        row = dataframe.apply(mutual_info_score, args=(dataframe[col],))
        rows.append(row)

    return pd.DataFrame(rows, index=dataframe.columns)



def compute_corrcoef(dataframe, threshold=0):
    connections = {}

    for neuron_pair in itertools.combinations(dataframe.columns, 2):
        if dataframe[neuron_pair[0]].std() == 0 or dataframe[neuron_pair[1]].std() == 0:
            continue

        corrcoef = np.corrcoef(dataframe[neuron_pair[0]].values, dataframe[neuron_pair[1]].values)[0][1]
        if corrcoef >= threshold:
            connections[neuron_pair] = corrcoef

    return connections

def affinity_propagation(similiarity_matrix):
    """Perform Affinity Propagation Clustering of data

    Note: This function is a wrapper for AffinityPropagation from scikit-learn
    Source: http://scikit-learn.org/stable/modules/generated/sklearn.cluster.AffinityPropagation.html

    Args:
        similarity_matrix: DataFrame, shape (n_samples, n_samples)
            Matrix of similarities between points.

    Returns:
        clusters: dictionary
            A dictionary of <sample: cluster label> key-value pairs.
    """
    clusters = {}

    labels = cluster.AffinityPropagation().fit_predict(similiarity_matrix)
    n_labels = labels.max()

    clusters = {}
    for i in range(n_labels+1):
        for neuron in list(similiarity_matrix.columns[labels==i]):
            clusters[neuron] = i

    return clusters

def extract_clusters(clusters):
    """Extract all the clusters into a dictionary of lists.

    This function takes in a dictionary of <neuron: cluster label>
    key-value pairs and produces a dictionary of <cluster label: neurons>

    Args:
        clusters: dictionary
            A dictionary of <sample: cluster label> key-value pairs.

    Returns:
        extracted_clusters: dictionary
            A dictionary of <cluster label: neurons list> key-value pairs.
    """
    extracted_clusters = {}

    for neuron, cluster_label in clusters.items():
        if extracted_clusters.get(cluster_label, None):
            extracted_clusters[cluster_label].append(neuron)
        else:
            extracted_clusters[cluster_label] = [neuron]

    return extracted_clusters



import os
import platform
import socket
import threading
import sys

class Server:

    def __init__(self, host, port, q):

        if platform.system() == "Windows":

            # Create a TCP/IP socket
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # Bind the socket to the port
            server_address = (host, port)
            print("Server starting up on {} port {}".format(server_address[0], server_address[1]))
            self.sock.bind(server_address)

        else:
            server_address = "./uds_socket"

            # Make sure the socket does not already exist
            try:
                os.unlink(server_address)
            except OSError:
                if os.path.exists(server_address):
                    raise

            # Create a UDS socket
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            # Bind the socket to the address
            print("Starting up on {}".format(server_address))
            self.sock.bind(server_address)

        # Listen for incoming connections
        self.sock.listen(5)

        self.q = q
        self.client_threads = {}
        listener_thread = threading.Thread(target=self.listen_for_clients, args=())
        listener_thread.daemon = True
        listener_thread.start()

    def listen_for_clients(self):
        print("Listening for clients...")

        while True:
            client, addr = self.sock.accept()
            print("Accepted Connection from: {}".format(client))
            t = threading.Thread(target=self.data_sender, args=())
            self.client_threads[client] = (t, addr)
            t.daemon = True
            t.start()

    def data_sender(self):
        while True:
            try:
                while True:
                    data = "{},".format(self.q.get())
                    for client, _ in self.client_threads.items():
                        client.sendall(data.encode())
            except (BrokenPipeError, socket.error) as e:
                print(e, file=sys.stderr)
                print("Connection to client: {} was broken!".format(client), file=sys.stderr)
                print(self.client_threads, file=sys.stderr)
                client.close()
                del self.client_threads[client]


#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-

"""
Helper functions for manipulating data from CNMFE (https://github.com/zhoupc/CNMF_E)

Many of these functions were written by Tim Machado (https://github.com/zhoupc/CNMF_E/tree/master/python_wrapper), or
were taken from Caiman (https://github.com/flatironinstitute/CaImAn)
"""

import os
import sys

sys.path.append("/OASIS/")

from matplotlib import pyplot as plt
import scipy.sparse as sparse
import numpy as np
import scipy.io as sio
import h5py
import pandas as pd
import analysis.analysis_utils as au

from oasis.functions import estimate_parameters
from oasis.oasis_methods import oasisAR1

def process_cnmfe(neuron, analysis_FR, normalize='std_noise', neural_FR=None, smin=10):
    """
    Deconvolves output from CNMFE and (optionally) normalizes the traces. The 
    raw traces and deconvolved spikes and traces are returned as TxN pandas 
    DataFrame

    Parameters
    ----------
    neuron : dict
        CNMFE results
    analysis_FR : int
        Frames per second that the neural data will be resampledto
    normalize : str
        Indicates which type of normalization to apply to traces
            ('zscore', 'std_noise'). Any other input will prevent normalization
    neural_FR : TYPE, optional
        The current frame rate of the calcium traces. If none is provided,
        this value will be taken from neuron['Fs']. The default is None.
    smin : int, optional
        Minimum spike size is smin*sn, where sn is the computed noise for
        each neuron. The default is 10.

    Returns
    -------
    f : pandas DataFrame
        raw traces
    s : pandas DataFrame
        deconvolved spikes
    c : pandas DataFrame
        deconvolved traces
    """


    F = pd.DataFrame(neuron['C_raw'].T)
    C = pd.DataFrame(neuron['C'].T)
    S = pd.DataFrame(neuron['S'].T)

    #Add timedelta column as index of neural DFs
    if neural_FR is None:
        neural_FR = neuron['Fs']
    frame_length = str(np.round(1000/neural_FR,3)) + 'ms'
    analysis_frame_length = str(np.round(1000/analysis_FR,3)) + 'ms'

    f = au.resample_ts(F, frame_length, analysis_frame_length)
    c = au.resample_ts(C, frame_length, analysis_frame_length)
    s = au.resample_ts(S, frame_length, analysis_frame_length)

    c,s,f = deconvolve_df(f, smin=smin, normalize = normalize) #normalize = 'zscore', 'std_noise'

    c = pd.DataFrame(c)
    s = pd.DataFrame(s)
    f = pd.DataFrame(f)

    s.columns = [x+1 for x in s.columns]
    c.columns = [x+1 for x in c.columns]
    f.columns = [x+1 for x in f.columns]

    return f, s, c
  
def import_cnmfe_caiman(file_path):
    """
    Reads in .hdf5 file containing the estimates structure from CaImAn CNMF-E
    
    Inputs:
        file_path: str
            File path for the .hdf5 CNMF-E CaImAn results
    
    Returns:
        neuron: dict
            Dictionary containing raw, c, s, and frame rate from the .hdf5 results file
    """
    
    # Loads a cnmfe results file that has been saved as a results.hdf5
    cnmf_obj = h5py.File(file_path, 'r')
    
    # Extracts the caiman estimates and frame rate of the calcium imaging recording
    data = cnmf_obj['estimates']
    fr = cnmf_obj['params']['data']['fr'][()]
    
    # Extracts raw, C, and S from the caiman estimates 
    raw = data['C'][()] + data['YrA'][()]
    C = data['C'][()]
    S = data['S'][()]
    
    # Constructs a neuron dictionary to interface with process_cnmfe function
    neuron = {'C_raw':raw,
              'C':C,
              'S':S,
              'Fs':fr}
    
    return neuron
  
def deconvolve_df(F, smin=10, normalize = None):
    """
    Deconvolve raw fluorescence traces

    Inputs:

        F: N x T pandas DataFrame
            Raw fluorescence traces for 1 mouse. This is neuron.C_raw

        smin: float
            Used to limit the minimim spike size. Min spike size = smin * sn

        normalize: str
            Options:
                - 'zscore': divide all traces by standard deviation of raw fluorescence trace.
                - 'std_noise': divide all traces by standard devation of the noise (C_raw - C)
                - anything else: don't normalize
    Returns:
        c: dict
            deconvolved trace
        s: dict
            deconvolved spikes
        F_norm: dict
            normalized raw trace

    """

    c=dict()
    s=dict()
    F_norm = dict()

    #make all nonnegative values 0 to ensure oasis convergence
    #F = F[F>0]
    #F = F.fillna(0)

    for i in F.columns:
        #define which trace to use
        f = F[i].values

        #estimate g, sn parameters
        est = estimate_parameters(f,p=1,fudge_factor=0.99)
        g=est[0]
        sn=est[1]
        # Deconvolve with threshold s_min = 10*sn

        c[i], s[i] = oasisAR1(f, g = g, s_min = smin*sn)


        #Optional: normalize f,c,s
        # from PC: instead of normalizing by dividing by std(F), normalize by dividing by std(F-C)
        sigma = 1
        if normalize == 'zscore':
            sigma = np.std(f)
        elif normalize == 'std_noise':
            sigma = np.std(f - c[i])

        f = f / sigma
        c[i] = c[i] / sigma
        s[i] = s[i] / sigma

        F_norm[i] = f

    return c, s, F_norm

def get_cnmfe_data(data_folder, mouse, behavior):
    """
    Reads in .mat file containing the neuron structure from Matlab CNMF_E
    Assumes folder structure is designed as follows:
        data_folder/mouse/behavior/cnmfe_results.mat
    Converts all sparse arrays to dense

    Inputs:
        data_folder: str
            Folder where all the data files are located
        mouse: str
            Name of the mouse
        behavior:str
            Name of the behavior
    Returns:
        data: dict
            Python dictionary containing all data from matlab
    """
    import glob

    mouse_folder = os.path.join(data_folder,mouse)

    if not os.path.isdir(mouse_folder):
        print('Error! There is no folder for ' + mouse)
        return

    behavior_folder = None
    for folder in os.listdir(mouse_folder):
        #if behavior in folder:
        if folder.endswith(behavior):
            behavior_folder = folder
    if not behavior_folder:
        print('Error! There is no folder for '+ mouse + ' '+ behavior)
        return

    file = glob.glob(os.path.join(mouse_folder, behavior_folder, '*.mat'))
    data={}
    if file:
        file = file[0]
        data = loadmat(file)

        #if neuron was saved using CNMFE's save function, need to load neuron_results
        if 'C_raw' not in data.keys():
            if 'neuron_results' in data.keys():
                data = data['neuron_results']
        #data = sio.loadmat(os.path.join(mouse_folder, behavior_folder, 'cnmfe_results.mat'), squeeze_me = True, struct_as_record = False)

    for k,v in data.items():
        if sparse.issparse(v):
            data[k] = v.todense()
    return data

def loadmat(filename):
    '''
    this function should be called instead of direct sio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects

    from https://stackoverflow.com/questions/7008608/scipy-io-loadmat-nested-structures-i-e-dictionaries/7418519
    '''
    data = sio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)

def _check_keys(dict):
    '''
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries

    from https://stackoverflow.com/questions/7008608/scipy-io-loadmat-nested-structures-i-e-dictionaries/7418519

    '''
    for key in dict:
        if isinstance(dict[key], sio.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict

def _todict(matobj):
    '''
    A recursive function which constructs from matobjects nested dictionaries

    from https://stackoverflow.com/questions/7008608/scipy-io-loadmat-nested-structures-i-e-dictionaries/7418519

    '''
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, sio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict

def xy_dist(p1, p2):
    """
    Computes the euclidian distance between two 2D points

    Inputs:
        p1,p2: tuple of floats
            The (x,y) location of each point
    Returns:
        dist: float
            The Euclidian distance between p1 and p2
    """

    dist = np.sqrt((p1[0]-p2[0])**2 + (p1[1] - p2[1])**2)
    return dist

def normalize(trace, percentile=True):
    """ Normalize a fluorescence trace by its max or its 99th percentile. """
    #trace = trace - np.min(trace)

    if percentile:
        if np.percentile(trace, 99) > 0:
            trace = trace / np.percentile(trace, 99)

    elif np.max(trace) > 0:
        trace = trace / np.max(trace)

    return trace


def com(A, d1, d2):
    """Calculation of the center of mass for spatial components

       From Caiman: https://github.com/flatironinstitute/CaImAn
       @author: agiovann

     Inputs:
     ------
     A:   np.ndarray
          matrix of spatial components (d x K)

     d1:  int
          number of pixels in x-direction

     d2:  int
          number of pixels in y-direction

     Output:
     -------
     cm:  np.ndarray
          center of mass for spatial components (K x 2)
    """
    nr = np.shape(A)[-1]
    Coor = dict()
    Coor['x'] = np.kron(np.ones((d2, 1)), np.expand_dims(list(range(d1)),
                                                         axis=1))
    Coor['y'] = np.kron(np.expand_dims(list(range(d2)), axis=1),
                        np.ones((d1, 1)))
    cm = np.zeros((nr, 2))        # vector for center of mass
    cm[:, 0] = old_div(np.dot(Coor['x'].T, A), A.sum(axis=0))
    cm[:, 1] = old_div(np.dot(Coor['y'].T, A), A.sum(axis=0))

    return cm


def plot_contours(A, Cn, thr=None, thr_method='max', maxthr=0.2, nrgthr=0.9,
                  display_numbers=True, max_number=None,
                  cmap=None, swap_dim=False, colors='w', vmin=None,
                  vmax=None, neuron_list=None, show_bg=False, **kwargs):
    """Plots contour of spatial components against a background image
       and returns their coordinates

       From Caiman: https://github.com/flatironinstitute/CaImAn
       @author: agiovann

     Parameters:
     -----------
     A:   np.ndarray or sparse matrix
               Matrix of Spatial components (d x K)

     Cn:  np.ndarray (2D)
               Background image (e.g. mean, correlation)

     thr_method: [optional] string
              Method of thresholding:
                  'max' sets to zero pixels that have value less
                  than a fraction of the max value
                  'nrg' keeps the pixels that contribute up to a
                  specified fraction of the energy

     maxthr: [optional] scalar
                Threshold of max value

     nrgthr: [optional] scalar
                Threshold of energy

     thr: scalar between 0 and 1
               Energy threshold for computing contours (default 0.9)
               Kept for backwards compatibility.
               If not None then thr_method = 'nrg', and nrgthr = thr

     display_number:     Boolean
               Display number of ROIs if checked (default True)

     max_number:    int
               Display the number for only the first max_number components
               (default None, display all numbers)

     cmap:     string
               User specifies the colormap (default None, default colormap)

     neuron_list:   list
               User specififies which neurons to plot

     Returns:
     --------
     Coor: list of coordinates with center of mass,
        contour plot coordinates and bounding box for each component
    """

    linewidths = kwargs.get('linewidths', 1)
    fontsize = kwargs.get('fontsize', 12)
    fontname = kwargs.get('fontname', 'Arial')
    fontweight = kwargs.get('fontweight', 'bold')
    fontcolor = kwargs.get('fontcolor', colors)
    if sparse.issparse(A):
        A = np.array(A.todense())
    else:
        A = np.array(A)

    if swap_dim:
        Cn = Cn.T
        print('Swapping dim')

    d1, d2 = np.shape(Cn)
    d, nr = np.shape(A)
    if max_number is None:
        max_number = nr

    if neuron_list:
        neuron_list = [i-1 for i in neuron_list]
    else:
        neuron_list = range(np.minimum(nr, max_number))

    if thr is not None:
        thr_method = 'nrg'
        nrgthr = thr
        print("The way to call utilities.plot_contours has changed.")

    if show_bg:
        x, y = np.mgrid[0:d1:1, 0:d2:1]
    else:
        x, y = np.mgrid[d1:0:-1, 0:d2:1] #reverse y direction to get correct orientation
    ax = plt.gca()

    if show_bg:
        if vmax is None and vmin is None:
            plt.imshow(Cn, interpolation=None, cmap=cmap,
                       vmin=np.percentile(Cn[~np.isnan(Cn)], 1),
                       vmax=np.percentile(Cn[~np.isnan(Cn)], 99))
        else:
            plt.imshow(Cn, interpolation=None, cmap=cmap,
                       vmin=vmin, vmax=vmax)

    coordinates = []
    cm = com(A, d1, d2)
    for i in neuron_list:
        pars = dict(kwargs)
        if thr_method == 'nrg':
            indx = np.argsort(A[:, i], axis=None)[::-1]
            cumEn = np.cumsum(A[:, i].flatten()[indx]**2)
            cumEn /= cumEn[-1]
            Bvec = np.zeros(d)
            Bvec[indx] = cumEn
            thr = nrgthr

        else:
            if thr_method != 'max':
                print("Unknown threshold method. Choosing max")
            Bvec = A[:, i].flatten()
            Bvec /= np.max(Bvec)
            thr = maxthr

        if swap_dim:
            Bmat = np.reshape(Bvec, np.shape(Cn), order='C')
        else:
            Bmat = np.reshape(Bvec, np.shape(Cn), order='F')
        cs = plt.contour(y, x, Bmat, [thr], colors=colors, linewidths = linewidths)

        # this fix is necessary for having disjoint figures and borders
        p = cs.collections[0].get_paths()
        v = np.atleast_2d([np.nan, np.nan])
        for pths in p:
            vtx = pths.vertices
            num_close_coords = np.sum(np.isclose(vtx[0, :], vtx[-1, :]))
            if num_close_coords < 2:
                if num_close_coords == 0:
                    # case angle
                    newpt = np.round(old_div(vtx[-1, :], [d2, d1])) * [d2, d1]
                    vtx = np.concatenate((vtx, newpt[np.newaxis, :]), axis=0)

                else:
                    # case one is border
                    vtx = np.concatenate((vtx, vtx[0, np.newaxis]), axis=0)

            v = np.concatenate((v, vtx, np.atleast_2d([np.nan, np.nan])),
                               axis=0)

        pars['CoM'] = np.squeeze(cm[i, :])
        pars['coordinates'] = v
        pars['bbox'] = [np.floor(np.min(v[:, 1])), np.ceil(np.max(v[:, 1])),
                        np.floor(np.min(v[:, 0])), np.ceil(np.max(v[:, 0]))]
        pars['neuron_id'] = i + 1
        coordinates.append(pars)

    if display_numbers:
        for i in neuron_list:

            x = cm[i, 1]
            if show_bg:
                y = cm[i, 0]
            else:
                y = d1-cm[i, 0] #need to flip to preserve location if theres no background

            if swap_dim:
                ax.text(y, x, str(i + 1), color=colors)
            else:
                ax.text(x, y, str(i + 1), color=fontcolor,
                        fontsize=fontsize, fontname=fontname,
                        fontweight=fontweight)

    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.set_yticks([])
    ax.set_xticks([])
    ax.set_aspect('equal')
    
    return coordinates

def old_div(a, b):
    """
    DEPRECATED: import ``old_div`` from ``past.utils`` instead.

    Equivalent to ``a / b`` on Python 2 without ``from __future__ import
    division``.

    TODO: generalize this to other objects (like arrays etc.)
    """

    return a // b


"""
This module contains all the functions responsible for executing a
permutation test.

@author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""

import os
from multiprocessing import Process
from multiprocessing import Queue
import numpy as np
import pandas as pd

class Resampler:
    """
    This class is meant to be a toolbox for the purposes of executing
    permutation resampling, and in order to carry out permutation tests.
    """

    def __init__(self):
        pass

    @staticmethod
    def get_num_of_events(dataframe, neuron):
        """Get the number of signal spikes for a given column vector

        Args:
            dataframe: DataFrame

                A pandas DataFrame that contains at least one neuron's signal
                data, in column vector form.

            neuron:
                The name of the neuron column vector to get the number
                of events for.

        Returns:
            The amount of datapoints in a given column vector of nonzero
            value.
        """
        return len(dataframe.loc[:, neuron][dataframe[neuron] != 0])

    @staticmethod
    def diff_of_mean_rate(dataframe, *beh_col_vec, frame_rate=10):
        """Compute difference of means between the rates of two behaviors.

        Args:
            dataframe: DataFrame
                A pandas DataFrame of all the neuron column vectors
                for a given animal.

            beh_col_vec: pandas Series
                A single behavior column vector or two behavior column vectors
                used to compute the difference of means rate.
                e.g. "OpenArms" vs. "ClosedArms".

            frame_rate: int, optional, default: 10
                The framerate associated with the given data.

        Returns: numpy array
            A numpy array of all the difference of means, D_hat, values, i.e.,
            all of the behavior vectors means subtracted from the corresponding
            means of the non-behavior vectors, all scaled by the frame rate.
        """
        if len(beh_col_vec) != 1 and len(beh_col_vec) != 2:
            raise ValueError("You provided an appropriate amount of behaviors.")
        if len(beh_col_vec) == 1:
            beh_vec = dataframe.loc[beh_col_vec[0] != 0]
            no_beh_vec = dataframe.loc[beh_col_vec[0] == 0]
            return frame_rate * (beh_vec.values.mean(axis=0) - no_beh_vec.values.mean(axis=0))
        if len(beh_col_vec) == 2:
            beh_vec = dataframe.loc[beh_col_vec[0] != 0]
            no_beh_vec = dataframe.loc[beh_col_vec[1] != 0]
            return frame_rate * (beh_vec.values.mean(axis=0) - no_beh_vec.values.mean(axis=0))

    @staticmethod
    def __shuffle_worker(queue, resamples, dataframe, statistic, *beh_col_vec, flip_roll=False):
        """Helper function for shuffle()

        This function repeats the permutation resampling and computation of
        the test statistic (difference of means), a *resamples* amount of times.
        This allows us to create a permutation distribution for each neuron
        column vector, under the condition that the null hypothesis is true,
        i.e., that the neuron is not-selective for the behaviors.

        NOTE: This function is meant to be only be used as a helper function
        for the shuffle() function.

        Args:
            queue: Queue
                A thread-safe FIFO data structure to which the resulting
                dataframe will be added.

            resamples: int
                The amount of permutation resamples to draw from the data.

            dataframe: DataFrame
                The data to be used to randomly draw permutation resamples.

            statistic: function
                A function that will compute a statistic that measures the size
                of an effect of interest (e.g. difference of means, mutual
                information, and etc.)

            beh_col_vec: pandas Series
                The columns vectors to be used as the two groups to
                use for permutation resamples.

            flip_roll: boolean, optional
                If data should be flipped and then randomly rolled for each
                resample; default is false.
        """
        if isinstance(dataframe, pd.DataFrame):
            column_names = list(dataframe.columns)
        elif isinstance(dataframe, pd.Series):
            column_names = [1]

        rows_list = []
        for i in range(resamples):
            
            if flip_roll:
                dataframe = dataframe.reindex(np.roll(dataframe.index, np.random.randint(1, high=len(dataframe.index)+1)))
                dataframe.index = range(int(len(dataframe.index)))
            else:
                dataframe = dataframe.sample(frac=1)
                dataframe.index = range(int(len(dataframe.index)))
        
            row = statistic(dataframe, *beh_col_vec)
            if len(column_names) > 1:
                rows_list.append(dict(zip(column_names, row)))
            else:
                rows_list.append({column_names[0]: row})

        queue.put(pd.DataFrame(rows_list, columns=column_names))

    @staticmethod
    def shuffle(resamples, dataframe, statistic, *beh_col_vec, flip_roll=False):
        """Permutation resampling function for neuron selectivty analysis.

        This function simply starts a new process for each CPU that the machine
        has. More specifically, this function starts the shuffle_worker()
        function for each new process, in order to allow the permutation
        distribution for each neuron column vector to be created in a more
        expedited fashion. More specifically, the amount of required permutation
        resamples is split evenly amongst all of the CPU's of the machine this
        function will be run on.

        Args:
            resamples: int
                The total amount of permutation resamples desired.

            dataframe: DataFrame
                The data to be used to randomly draw permutation resamples.

            statistic: function
                A function that will compute a statistic that measures the size
                of an effect of interest (e.g. difference of means, mutual
                information, and etc.)

            beh_col_vec: pandas Series
                The columns vectors to be used as the two groups to
                use for permutation resamples.

            flip_roll: boolean, optional, default: False
                If data should be flipped and then randomly rolled for each
                resample.

        Returns:
            A (vertically) concatenated pandas DataFrame of all the DataFrames
            the shuffle_worker processes produced.
        """
        if flip_roll:
            dataframe=dataframe.reindex(np.flip(dataframe.index))
            dataframe.index = range(int(len(dataframe.index)))
                                    
        keywords = {"flip_roll": flip_roll}
        resamples_per_worker = resamples // os.cpu_count()
        queue = Queue()
        processes = []
        rets = []
        for i in range(os.cpu_count()):
            np.random.seed(i)
            process = Process(target=Resampler.__shuffle_worker, args=(queue, resamples_per_worker, dataframe, statistic, *beh_col_vec), kwargs=keywords)
            processes.append(process)
            process.start()
        for process in processes:
            ret = queue.get()  # will block
            rets.append(ret)
        for process in processes:
            process.join()
        
        #start columns at 1, not 0
        shuffle_dists = pd.concat(rets, ignore_index=True)
        
        return shuffle_dists

    @staticmethod
    def p_value(original_statistic, permutation_distribution):
        """Compute a two-sided p-value.

        This function is meant to compute a two-sided p-value on a given
        permutation distribution.

        Args:
            original_statistic: float
                The original value of the statistic computed on the data.

            permutation_distribution: DataFrame or Series
                A pandas DataFrame of the permutation distributions.

        Returns:
            p_val: float
                The p-value that was located/computed.

        """
        D = original_statistic
        D_i = permutation_distribution.abs()
        P = len(permutation_distribution.index)
        p_val = (1/P) * len(permutation_distribution.loc[D_i >= abs(D)])

        return p_val

    def z_score(original_statistic, permutation_distribution):
        """
        Compute z-score for a given value, and the permutation distribution
        
        Args:
            original_statistic: float
                The original value of the statistic computed on the data.

            permutation_distribution: DataFrame or Series
                A pandas Series of the permutation distributions.

        Returns:
            float
                The z-score that was computed.

        
        """
        mew = permutation_distribution.mean()
        std = permutation_distribution.std()
        
        return (original_statistic - mew) / std
        
    
    @staticmethod
    def two_tailed_test(original_statistic, permutation_distribution, **kwargs):
        """Conduct a two-tailed hypothesis test.

        WARNING: Use this function ONLY if your permutation distribution is
        normally distributed.

        Args:
            original_statistic: float
                The original value of the statistic computed on the data.

            permutation_distribution: DataFrame
                A pandas DataFrame of the permutation distributions.

            high: float, optional
                The cutoff for the upper-tail of the distribution.

            low: float, optional
                The cutoff for the lower-tail of the distribution.

        Returns:
            1 if the original statistic is observed in the upper-tail of the
            permutation distribution.
            -1 if the original statistic is observed in the lower-tail of the
            permutation distribution.
            0 if the original statistic is not observed in either tail of the
            permutation distribution.

        """
        high = kwargs.get("high", 95)
        low = kwargs.get("low", 5)

        if original_statistic >= np.percentile(permutation_distribution, high):
            return 1
        elif original_statistic <= np.percentile(permutation_distribution, low):
            return -1
        else:
            return 0



"""
This module contains the NeuronNetwork class, which is to be used for quickly
generating and plotting visualizations of neuron networks.

@authors: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
          Jack Berry, Columbia University, jeb2242@columbia.edu
"""

import sys
import itertools
import numpy as np
import networkx as nx
from networkx.algorithms.approximation import clique
from matplotlib.pylab import plt

class NeuronNetwork:
    """
    This is a wrapper class for the NetworkX graph data structure. Use this
    class in order to conduct graph theory analysis on networks of neurons
    that were observed during experiments/trials.
    """

    def __init__(self, neurons, connections):
        self.neurons = list(neurons)
        self.connections = connections
        self.network = self.create_graph(self.neurons, self.connections)

    def create_graph(self, nodes, edges):
        """Wrapper function for creating a NetworkX graph.

        Each individual column of the provided DataFrame will be represented by
        a single node in the graph. Each pair of correlated nodes (neurons)
        will be connected by an edge, where the edge will receive a weight of
        the specific correlation coefficient of those two nodes.

        Args:
            nodes: list
                A list of all the neurons/nodes in the network/graph.

            edges: dict {(int, int): scalar, ..., (int, int): scalar}
                A dictionary of key-value pairs, where each key is a tuple
                representing an edge between two neurons/nodes, and each value
                is a scalar value representing the weight of that edge.

        Returns:
            graph: NetworkX graph
                A graph of the neuronal network.
        """
        graph = nx.Graph()
        graph.add_nodes_from(nodes)

        for edge in edges:
            graph.add_edge(edge[0], edge[1], weight=edges.get(edge, None))

        return graph

    def plot(self, **kwargs):
        """A wrapper function for plotting a NetworkX graph

        This function will draw a provided NetworkX graph using either the
        spring layout algorithm, or by the positions provided.

        Args:
            pos: dict, optional, default: networkx.drawing.layout.spring_layout
                A dictionary of the network's neurons as keys and their (x, y)
                coordinates as corresponding values.

            node_size: int, optional
                The size of the plotted neurons in the network.

            node_colors: list, optional
                The colors of the neurons to be plotted.

            fontsize: int, optional, default: 10
                The size of the font labelling the plotted neurons.

            draw_edges: bool, optional, default: True
                When True, the edges between all nodes will be drawn. When
                False, the edges between all nodes will be omitted from the
                figure.

            save: bool, optional, default: False
                When True, the plotted figure will be saved to the current
                working directory with its title as the file name, in PDF
                format.

            title: str, optional, default: None
                The title of the plotted graph/network.

            a: float, optional
                The transparency of the nodes
                
            label_nodes: bool, optional
                Whether to display node labels. Default is True
                
            edge_a: float, optional
                The transparency of the edges. Default is 1
            
            edge_weight: float, optional
                The width of the edge. Default is 2
                
            label_edges: bool, optional
                Whether to display edge labels. Default is False.

        Returns:
            pos: dict
                A dictionary of the network's neurons as keys and their (x, y)
                coordinates as corresponding values.
        """

        # Get positions for all nodes
        pos = kwargs.get("pos", None)
        if pos is None:
            print("You did not provide a neuron position dictionary. The spring layout function will be used to plot the network", file=sys.stderr)
            pos = nx.spring_layout(self.network, weight="weight");

        # Size of the plot
        figsize = kwargs.get("figsize", (30, 30))
        plt.figure(figsize=figsize)

        # Nodes
        node_size = kwargs.get("node_size", 600)
        node_colors = kwargs.get("node_colors", self.neurons)
        a = kwargs.get('a',.5)
        disp_node_labels=kwargs.get('label_nodes',True)
        nx.draw_networkx_nodes(self.network, pos, alpha=a, node_size=node_size, cmap=plt.cm.Dark2, node_color=node_colors)

        _, weights = zip(*nx.get_edge_attributes(self.network, "weight").items())

        # Draw edges
        edgeweight=kwargs.get('edgeweight',2)
        edge_a = kwargs.get('edge_a',1)
        edge_color = kwargs.get('edgecolor',weights)
        edge_cmap = kwargs.get('edge_cmap',plt.cm.coolwarm)
        edgelist = kwargs.get('edgelist',self.network.edges)
        if kwargs.get("draw_edges", True):
            nx.draw_networkx_edges(self.network, pos, alpha=edge_a, width=edgeweight,edge_color=edge_color,edge_cmap=edge_cmap,edgelist=edgelist)
            sm = plt.cm.ScalarMappable(cmap=edge_cmap)
            sm._A = []
            #cbar = plt.colorbar(sm, shrink=0.2, aspect=10, anchor = (0,10))
  
        
        # Labels
        
        label_edges=kwargs.get("label_edges",False)
        if label_edges:
            font_size = kwargs.get("font_size", 10)
            nx.draw_networkx_labels(self.network, pos, font_size=font_size)

        if disp_node_labels:
            font_size = kwargs.get("font_size", 10)
            nx.draw_networkx_labels(self.network, pos, font_size=font_size)
        
        title = kwargs.get("title", None)
        plt.title(title)
        plt.axis("off")

        save_to_file = kwargs.get("save", False)
     
        if save_to_file:
            file_name=kwargs.get("file_name","Graph")
            plt.savefig(file_name, dpi=300)

        plt.show()
        return pos
    
    def get_neighbors(self,neurons=None):
        """
        Returns a list of neighbors of a list of nodes
        
        inputs:
        g: graph with which to search for neighbors
        neurons: list of nodes to find neighbors of
        
        outputs:
        neighbors: list of neighbors
        
        """
        
        neighbors = []
        for n in self.neurons:
            if n in neurons:
    
                try:
                    neighbors.extend(list(self.network.neighbors(n)))
                except KeyError:
                    neighbors = list(self.network.neighbors(n))
        neighbors = set(neighbors)
        return neighbors
    
    
    def compute_connection_density(self):
        """Computes the connection density of a network of neurons.

            Connection density is the actual number of edges in the graph as a
            proportion of the total number of possible edges and is the simplest
            estimator of the physical cost — for example, the energy or other
            resource requirements — of a network. (Bullmore et al. 2009)

            Total number of possible edges is: n(n-1)/2,
            where n is the number of nodes (neurons) in the graph.
        """
        num_of_neurons = len(self.neurons)
        possible_num_of_edges = (num_of_neurons * (num_of_neurons-1)) / 2
        return len(list(self.network.edges())) / possible_num_of_edges

    def mean_betw_cent(self, weight="weight"):
        """Computes the mean betweeness centrality of a network of neurons.

            The centrality of a node measures how many of the shortest paths
            between all other nodes pairs in the network pass through it. A node
            with high centrality is thus crucial to efficient communication.
            (Bullmore et. al. 2009)

            https://en.wikipedia.org/wiki/Betweenness_centrality
        """
        betw_centrality = nx.betweenness_centrality(self.network, weight=weight)
        return np.mean(list(betw_centrality.values()))

    def compute_mean_degree_cent(self):
        """Computes the mean degree centrality of a network of neurons.

            The centrality of a node measures how many of the shortest paths
            between all other nodes pairs in the network pass through it. A node
            with high centrality is thus crucial to efficient communication.
            (Bullmore et. al. 2009)

            https://en.wikipedia.org/wiki/Centrality#Degree_centrality
        """
        graph_centrality = nx.degree_centrality(self.network)
        return np.mean(list(graph_centrality.values()))

    def compute_mean_eigen_cent(self):
        """Computes the mean Eigenvector centrality of a network of neurons.

            The centrality of a node measures how many of the shortest paths
            between all other nodes pairs in the network pass through it. A node
            with high centrality is thus crucial to efficient communication.
            (Bullmore et. al. 2009)

            https://en.wikipedia.org/wiki/Eigenvector_centrality
        """
        graph_centrality = nx.eigenvector_centrality(self.network, weight="weight")
        return np.mean(list(graph_centrality.values()))

    def compute_mean_katz_cent(self):
        """Computes the mean Katz centrality of a network of neurons.

           The centrality of a node measures how many of the shortest paths
           between all other nodes pairs in the network pass through it. A node
           with high centrality is thus crucial to efficient communication.
           (Bullmore et. al. 2009)

           https://en.wikipedia.org/wiki/Katz_centrality
        """
        graph_centrality = nx.katz_centrality(self.network, weight="weight")
        return np.mean(list(graph_centrality.values()))

    def compute_mean_load_cent(self):
        """Computes the mean load centrality of a network of neurons.

            The centrality of a node measures how many of the shortest paths
            between all other nodes pairs in the network pass through it. A node
            with high centrality is thus crucial to efficient communication.
            (Bullmore et. al. 2009)
        """
        graph_centrality = nx.load_centrality(self.network, weight="weight")
        return np.mean(list(graph_centrality.values()))

    def compute_max_clique_size(self):
        """Computes the size of the maxiumum clique in the network of neurons.

            A maximum clique of a graph, G, is a clique, such that there is no
            clique with more vertices.

            https://en.wikipedia.org/wiki/Clique_(graph_theory)#Definitions
        """
        return len(clique.max_clique(self.network))

    def compute_mean_clique_size(self):
        """Computes the mean clique size in the network of neurons.

        Finds all cliques in an undirected graph (network), and computes the
        mean size of all those cliques.

        Returns:
            mean: float
                The mean clique size of the network of neurons.
        """
        all_cliques = nx.enumerate_all_cliques(self.network)

        size = 0
        running_sum = 0
        for cliq in all_cliques:
            size += 1
            running_sum += len(cliq)

        mean = running_sum / size
        return mean

    def avg_shortest_path_len(self, weight="weight"):
        """Computes the average path shortest path length.

        This function computes the average path length L, as the average
        length of the shortest path connecting any paid of nodes in a
        network.

        Args:
            weight: str or None, optional, default: None
                If None, every edge has weight/distance/cost 1. If a string, use
                this edge attribute as the edge weight. Any edge attribute not
                present defaults to 1.

        Returns:
            avg_shortest_path_len: float
                The average shortest path length in the network of neurons.

        """
        graph = self.network
        node_list = self.neurons
        shortest_path_lengths = []

        for node_pair in itertools.combinations(node_list, 2):
            source = node_pair[0]
            target = node_pair[1]
            if not nx.has_path(graph, source, target):
                continue

            shortest_path_lengths.append(nx.shortest_path_length(graph, source=source, target=target, weight=weight))

        avg_shortest_path_len = np.mean(shortest_path_lengths)
        return avg_shortest_path_len

    def small_worldness(self, weight="weight"):
        """Computes the small worldness of the neuron network.

        Returns:
            small_worldness: float
                The clustering coefficient divided by the average shortest path
                length of the neuron network.
        """
        cluster_coeff = nx.average_clustering(self.network, weight=weight)
        avg_shortest_path_len = self.avg_shortest_path_len(weight=weight)
        small_worldness = cluster_coeff / avg_shortest_path_len
        return small_worldness



# -*- coding: utf-8 -*-
"""
Functions for merging calcium imaging data from different sessions that have
been aligned using CellReg (https://github.com/zivlab/CellReg).

@author: Jack Berry
"""

import os
import glob
import copy
import h5py
import pandas as pd
import numpy as np


def read_cellreg(cellreg_folder, mousename, sessions):
    """
    Loads important data from the CellReg folder. Currently, the only object
    returned is the cell_to_index_map array that is stored in the cellRegistered
    struct.

    Inputs:
        cellreg_folder: str
            Path to the folder containing the output from CellReg. The folder
            structure is assumed to be of the following format:
                cellreg_folder/mouse_name/CellRegfolder_session1_session2/
            Note that the session names must be in the name of the CellReg output
            folder, and the order of the sessions indicates the order used by
            CellReg

        mousename: str
            Name of the mouse

        sessions: list of str
            List of the session names. These names must be in the CellReg output
            folder name.

    Returns:
        cell_df: pandas DataFrame
            DataFrame in which the index (named reg) indicates the registered
            neuron IDs, and the columns (named by session) indicate the original
            neuron IDs for that session.

    """

    folder = os.path.join(cellreg_folder, mousename)
    
    #Get the mouse subfolder containing all CellRegs for that mouse
    cell_regs = next(os.walk(folder))[1]
    
    #Get the CellReg folder for the specified sessions
    cell_reg=''
    for d in cell_regs:
        if all(s in d for s in sessions):
            cell_reg = d
    cellregpath = os.path.join(folder, cell_reg)

    #Read in the cell_to_index_map and rename the index and columns
    filename = glob.glob(os.path.join(cellregpath, 'cellRegistered*'))
    if not filename:
        return None
    f = h5py.File(filename[0], 'r')
    cell_df = pd.DataFrame(np.array(f['cell_registered_struct']['cell_to_index_map'])).T.astype(int)
    cell_df.columns = [s for s in sessions]
    cell_df.index = [x+1 for x in cell_df.index]
    cell_df.index.name = 'Reg'
    return cell_df

def session_to_reg(cell_df, session, neuron):
    """
    Translates the session neuron ID to the registered neuron ID

    Inputs:
        cell_df: pandas DataFrame
            Dataframe representing the cells_to_index_map, ie the return of
            read_CellReg
        session: str
            The session in which the neuron was identified
        neuron: int
            The neuron ID in the session

    Returns:
        n_reg: int
            The registered neuron ID
    """

    n_reg = cell_df.index[cell_df[session] == neuron].item()
    return n_reg

def reg_to_session(cell_df, session, neuron):
    """
    Translates the registered neuron ID to the session neuron ID

    Inputs:
        cell_df: pandas DataFrame
            Dataframe representing the cells_to_index_map, ie the return of
            read_CellReg
        session: str
            The session in which the neuron was identified
        neuron: int
            The registered neuron ID

    Returns:
        n_session: int
            The session neuron ID
    """
    n_session = cell_df.loc[neuron, session]
    return n_session

def reorder_cols(df):
     df1 = df.reindex(sorted(df.columns), axis=1)   
     return df1
     
def register_mouse(mouse, cell_df, session=None):
    """
    Creates a new mouse object in which neuron IDs are given as the registered
    ID instead of the session ID.

    Inputs:
        mouse: Mouse object

        cell_df: pandas DataFrame
            Dataframe representing the cells_to_index_map, ie the return of
            read_CellReg

        session: str
            The session of the mouse object. If none is given, the session
            stored in the mouse object (mouse.session) will be used

    """

    session_neurons = mouse.spikes.columns
    if session is None:
        session = mouse.session
        if session is None:
            print('Error: No session given')
            return None
    reg_n = [session_to_reg(cell_df, session, n) for n in session_neurons]
    reg_mouse = copy.deepcopy(mouse)
    reg_mouse.spikes.columns = reg_n
    reg_mouse.cell_transients.columns = reg_n
    reg_mouse.raw.columns = reg_n
    
    #add silent neurons for all neurons that were not identified in the session
    silent_neuron = np.zeros(len(mouse.spikes))
    silent_ids = cell_df.index[cell_df[session] == 0].tolist()
    silent_df = pd.DataFrame({k:silent_neuron for k in silent_ids})
    
    reg_mouse.spikes = pd.concat([reg_mouse.spikes, silent_df], axis=1)
    reg_mouse.cell_transients = pd.concat([reg_mouse.cell_transients, silent_df], axis=1)
    reg_mouse.raw = pd.concat([reg_mouse.raw, silent_df], axis=1)
    
    #sort columns so that they are in increasing order
    reg_mouse.spikes = reorder_cols(reg_mouse.spikes)
    reg_mouse.cell_transients= reorder_cols(reg_mouse.cell_transients)
    reg_mouse.raw = reorder_cols(reg_mouse.raw)
    
    reg_mouse.spikes_and_beh = pd.concat([reg_mouse.spikes, reg_mouse.behavior],axis=1)

    return reg_mouse



"""
This module contains the classes for receiving, extracting, and plotting data.

@author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""

import os
import sys
import platform
import socket
import queue
import threading

import numpy as np
import pandas as pd
from PyQt5 import QtWidgets
import pyqtgraph as pg
from pyqt_wrapper import MainWindow
from data_dialogs import DataDialog, ColorsDialog

class DataGen:
    """Data Generator/Extractor class"""

    def __init__(self):
        self.filename = None
        self.show_file_dialog()
        if not self.filename[0]:
            sys.exit(1)

        try:
            self.dataset = pd.read_csv(self.filename[0])
        except (UnicodeDecodeError, pd.errors.ParserError) as exception:
            print(exception, file=sys.stderr)
            app = QtWidgets.QApplication([])
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Error")
            msg.setInformativeText(str(exception) + "\nMake sure you choose a csv file! Try again.")
            msg.setWindowTitle("Error")
            msg.show()
            app.exec_()
            sys.exit(1)

        # Prompt user to select neurons. If the user cancels the dialog, or the
        # the user closes the dialog without selecting anything, exit.
        self.neurons = self.show_data_dialog()
        if not self.neurons:
            sys.exit(1)

        # Prompt user to select behaviors. If the user cancels or closes the
        # dialogs, then it's assumed no behaviors were chosen/needed.
        self.behaviors = self.show_data_dialog(checkbox=True)

        # If behaviors of interest were selected, then create a dictionary with
        # each behavior coupled with a corresponding color.
        if self.behaviors:

            # If user checked box to choose custom colors, display the dialog
            # for the user to choose custom colors for each behavior. Finally,
            # ammend each color with a transparency value. Note that if the user
            # doesn't choose custom colors, a random color will be assigned to
            # each behavior.
            if self.choose_colors:
                self.behaviors = self.show_behavior_colors_dialog()
                for behavior, color in self.behaviors.items():
                    self.behaviors[behavior] = color
            else:
                temp = {}
                for behavior in self.behaviors:
                    rgb = list(np.random.randint(256, size=3))
                    rgb.append(40)
                    rgba = tuple(rgb)
                    temp[behavior] = rgba

                self.behaviors = temp

        self.dataset.fillna(0)
        self.neuron_col_vectors = self.dataset[self.neurons]

        # Make sure the user actually chose colors
        if self.behaviors:
            self.behavior_intervals = self.get_behavior(self.dataset)
        else:
            self.behavior_intervals = None

        # We no longer need the dataframe
        del self.dataset

    def show_file_dialog(self):
        """Display a file dialog to the user & let user choose a data file.
        """
        app = QtWidgets.QApplication(sys.argv)
        file_dialog = QtWidgets.QFileDialog()
        self.filename = file_dialog.getOpenFileName(None, "Choose a Data File", os.path.expanduser('~'))

    def show_data_dialog(self, checkbox=False):
        """Display a dialog to the user & let user choose neurons or behaviors.
        """
        app = QtWidgets.QApplication(sys.argv)
        data_dialog = DataDialog(list(self.dataset.columns), checkbox=checkbox)
        data_dialog.show()
        app.exec_()

        # If checkbox was added to dialog for choosing behaviors, then show
        # dialog for user to choose color for each behavior.
        if checkbox:
            self.choose_colors = data_dialog.choose_colors

        return data_dialog.selected_items

    def show_behavior_colors_dialog(self):
        """Display the behavior dialog to the user & let user choose colors.
        """
        app = QtWidgets.QApplication(sys.argv)
        colors_dialog = ColorsDialog(self.behaviors)
        colors_dialog.show()
        app.exec_()
        return colors_dialog.behavior_colors

    def get_behavior(self, dataset):
        all_behavior_intervals = []

        for behavior, color in self.behaviors.items():
            curr_beh_epochs = self.extract_epochs(dataset, behavior)
            curr_beh_intervals = self.filter_epochs(curr_beh_epochs[1], framerate=1, seconds=1)
            all_behavior_intervals.append((curr_beh_intervals, behavior, color))

        return all_behavior_intervals

    def extract_epochs(self, dataset, behavior):
        dataframe = dataset
        dataframe["block"] = (dataframe[behavior].shift(1) != dataframe[behavior]).astype(int).cumsum()
        df = dataframe.reset_index().groupby([behavior, "block"])["index"].apply(np.array)
        return df

    def filter_epochs(self, interval_series, framerate=10, seconds=1):
        return [interval for interval in interval_series if len(interval) >= framerate*seconds]

    def get_neuron_plots(self):
        return [self.neuron_col_vectors[col].values for col in self.neuron_col_vectors]


class Client:
    """Data receiver client"""

    def __init__(self, address, port, data_queue):
        self.data_queue = data_queue

        if platform.system() == "Windows":

            # Create a TCP/IP socket
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            # Connect the socket to the port where the server is listening
            server_address = (address, port)
            print("Connecting to {} port {}".format(server_address[0], server_address[1]))
            self.sock.connect(server_address)
        else:

            # Create a UDS socket
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            # Connect the socket to the port where the server is listening
            server_address = "./uds_socket"
            print("New client connecting to {}".format(server_address))

            try:
                self.sock.connect(server_address)
            except socket.error as msg:
                print(msg, file=sys.stderr)
                sys.exit(1)

        thread = threading.Thread(target=self.data_receiver, args=())
        thread.daemon = True
        thread.start()

    def data_receiver(self):
        print("New data receiver thread started...")
        try:
            while True:
                data = self.sock.recv(4096)
                if data:
                    data = data.decode()

                    for char in data.split(','):
                        if char:
                            if char == 'd':
                                self.data_queue.queue.clear()
                            else:
                                self.data_queue.put(char)
        except:
            print("Closing socket: {}".format(self.sock), file=sys.stderr)
            self.sock.close()
            return


def main():
    datagen = DataGen()
    plots = datagen.get_neuron_plots()
    plot_names = datagen.neurons

    data_queue = queue.Queue()

    # Create new plot window
    app = QtWidgets.QApplication(sys.argv)
    pg.setConfigOptions(antialias=True)  # set to True for higher quality plots
    main_window = MainWindow(data_queue, plots, plot_names, beh_intervals=datagen.behavior_intervals)
    main_window.show()
    main_window.resize(800, 600)
    main_window.raise_()
    _ = Client("localhost", 10000, data_queue)
    sys.exit(app.exec_())


if __name__ == "__main__":
    pg.setConfigOption("background", (230, 230, 230))
    pg.setConfigOption("foreground", (30, 30, 30))
    main()
